"""
Expose your connectors here so they are all importable from cranial.connectors.
"""
from cranial.connectors.base import Connector
from cranial.connectors.local import Connector as LocalConnector
from cranial.connectors.s3 import InMemoryConnector as S3InMemoryConnector



import logging
import io
import os

from google.cloud.storage import Client, Blob
from google.oauth2 import service_account

from cranial.connectors import base
from cranial.common import logger

log = logger.get('GOOGLECLOUDSTORAGE_LOGLEVEL', name='gcs_connector')


class InMemoryConnector(base.Connector):
    def __init__(self, bucket, prefix='', binary=True, do_read=False,
                 credentials=None, project=None):
        super().__init__(base_address=prefix, binary=binary, do_read=do_read)

        params = {'project': project}
        if credentials:
            creds = service_account.Credentials.from_service_account_info(
                credentials)
            params['credentials'] = creds
        self.bucket = Client(**params).get_bucket(bucket)
        self.bucket_name = bucket

    def get(self, name=None):
        """
        Get a bytes stream of an object

        Parameters
        ----------
        name
            if None, prefix will be used as a key name, otherwise name will be added to prefix after '/' separator

        Returns
        -------
            bytes stream with key's data, or an empty stream if errors occurred
        """
        key = self.base_address if name is None else os.path.join(self.base_address, name)
        key = key.strip('/')

        try:
            blob = Blob(key, self.bucket)
            filestream = io.BytesIO()
            blob.download_to_file(filestream)
            filestream.seek(0)
            log.info("Getting {} bytes from \tbucket={}\tkey={}".format(
                blob.size, self.bucket_name, key))
            self.metadata = blob.metadata
        except Exception as e:
            log.error("{}\tbucket={}\tkey={}".format(e, self.bucket_name, key))
            if log.level == logging.DEBUG:
                raise e

        if self.do_read:
            asbytes = filestream.read()

            if not self.binary:
                return asbytes.decode()
            else:
                return asbytes

        return filestream

    def put(self, source, name=None):
        """

        Parameters
        ----------
        source
        name

        Returns
        -------

        """

        if isinstance(source, io.BytesIO):
            filebuff = io.BufferedReader(source)
        elif isinstance(source, (str, bytes)):
            filebuff = io.BufferedReader(io.BytesIO(source))
        else:
            log.error('Source should be either a string, or bytes or io.BytesIO, got {}'.format(type(source)))
            return False

        key = self.base_address if name is None else os.path.join(
            self.base_address, name)
        key = key.strip('/')
        try:
            blob = Blob(key, self.bucket)
            blob.upload_from_file(filebuff, rewind=True)
            log.info("Uploaded {} bytes to \tbucket={}\tkey={}".format(
                len(source), self.bucket_name, key))
            return True
        except Exception as e:
            log.error("{}\tbucket={}\tkey={}".format(e, self.bucket, key))
            return False



from cassandra.encoder import Encoder
from cranial.common import logger

log = logger.get()


class Adapter():
    def __init__(self, session):
        self.session = session
        self.res = None
        # Use Cassandra Encoder to prevent SQL Injection
        self.encoder = Encoder()

    def sanitize(self, q, params):
        """
        Basic prevention against SQL Injection.
        >>> adapter = CassandraAdapter(None)
        >>> adapter.sanitize('SQL {}', ["injection'; DROP DATABASE;--"])
        "SQL 'injection''; DROP DATABASE;--'"
        """
        sanitized = map(self.encoder.cql_encode_all_types, params)
        return q.replace('?', '{}').format(*list(sanitized))

    def execute(self, q, params=None, executor=None):
        if executor is None:
            executor = self.session.execute

        try:
            if type(q) is str and params is not None and len(params):
                q = self.sanitize(q, params)
                self.res = executor(q)
            else:
                self.res = executor(q, params)
        except Exception as e:
            log.error('Exception {} on {}; {}'.format(e, q, str(params)))
            raise e

        self.rowcount = len(self.res.current_rows) if hasattr(self.res, 'current_rows') else None
        columns = getattr(self.res, 'column_names', [])
        self.description = [[c] for c in columns] if columns is not None else []
        return self.res

    def execute_async(self, q, params=None):
        return self.execute(q, params, executor=self.session.execute_async)

    def get_async(self):
        # Block for Async results...
        if hasattr(self.res, 'result'):
            self.res = self.res.result()
            self.rowcount = len(self.res.current_rows)
            columns = getattr(self.res, 'column_names', [])
            self.description = [[c] for c in columns] if columns is not None else []


    def fetchone(self):
        self.get_async()
        if self.rowcount:
            return self.res[0]
        else:
            # @ToDo This seems to assume KeyValueDB. We should generalize.
            raise KeyError('Key not found.')

    def fetchall(self):
        self.get_async()
        return self.res

    def __getattr__(self, name):
        """ Proxy anything else to the session object, for cases where DBAPI2
        compatible things already exist."""
        return getattr(self.session, name)


# Alias
CassandraAdapter = Adapter



'''
Helper functions to work with Cassandra via KeyValueDB.
'''
import json
import os

from cassandra.cluster import Cluster
from cassandra.policies import TokenAwarePolicy, RoundRobinPolicy
from cassandra.query import named_tuple_factory
from cassandra.auth import PlainTextAuthProvider

from cranial.datastore.adapters.cassandra import CassandraAdapter as Adapter
from cranial.keyvalue.keyvalue import KeyValueDB as KeyValue
from cranial.common import logger

log = logger.get()
session = None


def get_cursor(existing=None, hosts=None, new=False):
    global session

    if existing:
        return existing

    if session and not new:
        return session

    if hosts is None:
        hosts = os.environ.get('KEYVALUE_HOSTS', '').split(',')

    auth = None
    try:
        with open('keys/cassandra-config.json') as f:
            auth = PlainTextAuthProvider(**json.load(f))
    except Exception as e:
        log.warning('Could not get Cassandra credentials from file.')
        log.warning(e)

    cluster = Cluster(
        hosts,
        load_balancing_policy=TokenAwarePolicy(RoundRobinPolicy()),
        auth_provider=auth)

    session = cluster.connect('default')
    session.row_factory = named_tuple_factory

    return session


def get_kv(name, **kwargs):
    return KeyValue(Adapter(get_cursor()),
                    table=name,
                    create_table=True,
                    **kwargs)



from requests_futures.sessions import FuturesSession

from cranial.connectors import base


class Connector(base.Connector):
    r"""Connector for getting data via HTTP. Upload not yet implemented.

    >>> stream = Connector().get('http://httpbin.org/html')
    >>> next(stream).startswith('<!DOCTYPE html>')
    True
    >>> r = Connector('http://httpbin.org').getMultiple({'rabbit': 'image/png', 'grenade': 'bytes/10'}, binary=True)
    >>> next(r['rabbit'])
    b'\x89PNG\r\n'
    """

    def __init__(self, host: str = None):
        self.session = FuturesSession()
        # Ensure host ends in '/' if given.
        self.url_host = host if host is None or host.endswith('/') else host+'/'


    def get(self, url, binary=False):
        url = url if not self.url_host else self.url_host + url
        r = self.getFuture(url).result()
        return self.toStream(r.content if binary else r.text)


    def getFuture(self, url):
        url = url if not self.url_host else self.url_host + url
        return self.session.get(url)


    def getMultiple(self, d, blocking=True, binary=False):
        result = super(Connector, self).getMultiple(d, blocking)
        if not(blocking):
            return result

        for key in result:
            r = result[key]
            result[key] = self.toStream(r.content if binary else r.text)

        return result


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import importlib

class Param:
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return 'Param(%r)' % (self.value,)

def to_qmark(chunks):
    query_parts = []
    params = []
    for chunk in chunks:
        if isinstance(chunk, Param):
            params.append(chunk.value)
            query_parts.append('?')
        else:
            query_parts.append(chunk)
    return ''.join(query_parts), params

def to_numeric(chunks):
    query_parts = []
    params = []
    for chunk in chunks:
        if isinstance(chunk, Param):
            params.append(chunk.value)
            query_parts.append(':%d' % len(params))
        else:
            query_parts.append(chunk)
    return ''.join(query_parts), tuple(params)  # DCOracle2 has broken support
                                                # for sequences of other types
def to_named(chunks):
    query_parts = []
    params = {}
    for chunk in chunks:
        if isinstance(chunk, Param):
            name = 'p%d' % len(params)  # Are numbers in name allowed?
            params[name] = chunk.value
            query_parts.append(':%s' % name)
        else:
            query_parts.append(chunk)
    return ''.join(query_parts), params

def to_format(chunks):
    query_parts = []
    params = []
    for chunk in chunks:
        if isinstance(chunk, Param):
            params.append(chunk.value)
            query_parts.append('%s')
        else:
            query_parts.append(chunk.replace('%', '%%'))
    return ''.join(query_parts), params

def to_pyformat(chunks):
    query_parts = []
    params = {}
    for chunk in chunks:
        if isinstance(chunk, Param):
            name = '%d' % len(params)
            params[name] = chunk.value
            query_parts.append('%%(%s)s' % name)
        else:
            query_parts.append(chunk.replace('%', '%%'))
    return ''.join(query_parts), params


def get_paramstyle(conn):
    """ This should work with a connection or cursor object."""
    name = conn.__class__.__module__.split('.')[0]
    mod = importlib.import_module(name)
    return mod.paramstyle


def render_params(conn, chunks):
    try:
        style = get_paramstyle(conn)
    except:
        style = 'format'

    return globals()['to_' + style](chunks)


def get_temp_db(filename=None):
    import sqlite3
    if filename is None:
        conn = sqlite3.connect(":memory:")
    else:
        conn = sqlite3.connect(filename)
    return conn.cursor()



"""
A Fetcher is a the combination of a Connector and a Parser, providing a
standard way to retrieve data records as Dictionaries from arbitrary storage
systems, via a generator.
"""
from cranial.connectors.base import Connector
from cranial.parsers import base, line


class Parser():
    pass


class Fetcher():
    def __init__(self, connector: Connector, parser: base.Parser=line.Parser) -> None:
        self.connector = connector
        self.parser = parser

    def __iter__(self):
        return self.generator()

    def generator(self):
        for record in self.parser(self.connector):
            yield record



import io
import os
from tempfile import mkstemp
from concurrent import futures
from concurrent.futures import Future, ThreadPoolExecutor

from cranial.parsers import line


class Connector():
    def __init__(self, base_address='', binary=True, do_read=False):
        self.base_address = base_address
        self.binary = binary
        self.do_read = do_read

    def iterator(self, target: Any) -> iter:
        """Get one or more IOstreams and iterate over them yielding lines."""
        if type(target) == list:
            return self.generate_from_list(target)
        else:
            f = self.getFuture(target)
            return self.generate(f)

    def generate(self, future_item):
        for l in line.Parser(f.result()):
            yield l

    def generate_from_list(self, l: List):
        # @ToDo
        #getMultiple on all items, start returning whatever item arrives first.
        done = set()
        while len(done) < len(d):
            for key in d:
                if not (key in done):
                    try:
                        data = response[key].result(1)
                        for l in line.Parser(data):
                            yield l
                        done.add(key)
                    except futures.TimeoutError:
                        pass
                    except futures.CancelledError:
                        response[key] = False
                        done.add(key)
        return response


    def toStream(self, data):
        if type(data) is str:
            return io.StringIO(data)
        elif type(data) is bytes:
            return io.BytesIO(data)
        else:
            raise Exception('Not implemented for data type: {}'.format(
                type(data)))

    def get_tmp_file(self):
        fd, local_path = mkstemp()
        os.close(fd)
        return fd, local_path

    def get(self, name=None) -> io.BytesIO:
        """Return an IOstream."""
        raise Exception('Not Implemented')

    def put(self, stream: io.BytesIO, name: str = None) -> bool:
        """Return True if successful."""
        raise Exception('Not Implemented')

    def executor(self):
        if not hasattr(self, 'pool'):
            self.pool = ThreadPoolExecutor()
        return self.pool

    def _doFuture(self, fn: callable, *args, **kwargs) -> Future:
        return self.executor().submit(fn, *args, **kwargs)

    def getFuture(self, *args, **kwargs) -> Future:
        return self._doFuture(self.get, *args, **kwargs)

    def putFuture(self, *args, **kwargs) -> Future:
        return self._doFuture(self.put, *args, **kwargs)

    def _doMultiple(self, fn: callable, d: dict, blocking=True, **kwargs):
        """Takes a dict of keys to argument to pass to fn.

        Returns dict of Futures if 'blocking' is set False. Otherwise IOStreams.
        """
        response = {}
        for key in d:
            args = d[key]
            # Gracefully accept single arguments as-is.
            if type(args) is not list:
                args = [args]
            response[key] = fn(*args, **kwargs)

        if not blocking:
            return response

        done = set()
        while len(done) < len(d):
            for key in d:
                if not (key in done):
                    try:
                        data = response[key].result(1)
                        response[key] = data
                        done.add(key)
                    except futures.TimeoutError:
                        pass
                    except futures.CancelledError:
                        response[key] = False
                        done.add(key)
        return response

    def getMultiple(self, d: dict, blocking=True, **kwargs):
        return self._doMultiple(self.getFuture, d, blocking, **kwargs)

    def putMultiple(self, d: dict, blocking=True, **kwargs):
        return self._doMultiple(self.putFuture, d, blocking, **kwargs)

    def __str__(self):
        ss = [str(self.__class__).split("'")[1]]
        for attr in dir(self):
            if (not attr.startswith("_")) and (not hasattr(getattr(self, attr), '__call__')):
                ss.append("{} = {}".format(attr, getattr(self, attr))[:100])
        return '\t'.join(ss)



import io
import logging
import os
import shutil
import subprocess
import tempfile
from typing import Dict

import boto3

from cranial.connectors import base
from cranial.common import logger

log = logger.get('S3_LOGLEVEL', name='s3_fetchers')

SOURCE_DIR = 'storage/source'
TARGET_DIR = 'storage/target'
MODEL_DIR = 'storage/model'


def cleanup_temp_data():
    """
    delete temp dirs
    """
    try:
        shutil.rmtree(SOURCE_DIR)
        log.info("Removed {}".format(SOURCE_DIR))
    except Exception as e:
        log.info(e)
    try:
        shutil.rmtree(TARGET_DIR)
        log.info("Removed {}".format(TARGET_DIR))
    except Exception as e:
        log.info(e)
    try:
        shutil.rmtree(MODEL_DIR)
        log.info("Removed {}".format(MODEL_DIR))
    except Exception as e:
        log.info(e)


def prepare_s3_prefix(bucket, pre_date_part, y=None, mo=None, d=None, h=None, after_date_part=''):
    """
    compose an s3 path from given parts and date values
    example:
        {bucket}/{pre_date_part}/mo={y}-{mo}/day={d}/hour={h}/after_date_part
    Parameters
    ----------
    bucket
        bucket name, may include "s3://"
    pre_date_part
        part of the path before date strings
    y
        year number
    mo
        month number
    d
        day number
    h
        hour number
    after_date_part
        part of the path after date strings
    Returns
    -------
        composed s3 path
    """
    if not bucket.startswith('s3://'): bucket = 's3://' + bucket
    s = '/'.join([bucket.strip('/'), pre_date_part.strip('/')])
    if y is not None and mo is not None:
        s += '/mo={y}-{mo}'.format(y=y, mo=mo)
    if d is not None:
        s += '/day={d}'.format(d=d)
    if h is not None:
        s += '/hour={h}'.format(h=h)
    if len(after_date_part) > 0:
        s += '/' + after_date_part

    return s


def read_key(key, bucket, decode=True, verbose=False):
    """
    The intended use is within a map function, so to change bucket and verbosity is not extremely simple and needs
    some functools (partial?)

    Parameters - self explanatory

    Returns - contents of the s3 key, decoded into utf8 string
    """
    if verbose:
        log.info('reading {}'.format(key))
    try:
        b_str = boto3.resource('s3').Bucket(bucket).Object(key).get()['Body'].read()
    except Exception as e:
        log.error("{}:{}\t{}".format(bucket, key, e))
        return None
    return b_str.decode() if decode else b_str


def key_download_decompress(bucket, key, force_decompress=False):
    """
    First step in getting events from s3 key:
        - download a key into a temp file
        - decompress on disk
    Returns
    -------
        name of the decompressed temp file or None if failed
    """
    fp = tempfile.mkstemp(suffix='.gz')[1]
    boto3.resource('s3').Bucket(bucket).download_file(key, fp)
    log.info("downloaded {}/{} to {}".format(bucket, key, fp))
    if key.endswith('.gz') or force_decompress:
        # @TODO Replace with gzip module.
        p = subprocess.run(['gunzip', fp, '-f'], stderr=subprocess.PIPE)

        if p.returncode == 0:
            log.info('decompressed {}'.format(fp))
            return fp[:-3]
        else:
            log.warning(p.stderr)
            return None


class S3Connector(base.Connector):
    def __init__(self, bucket: str, prefix='') -> None:
        self.bucket = bucket
        self.prefix = prefix
        self.cache = {}  # type: Dict[str, str]
        log.info('s3.S3Connector downloads whole files to local disk '
                 + 'before use. Consider s3.InMemoryConnector instead.')

    def get(self, s3_path, binary=False, redownload=False):
        if not redownload:
            try:
                local_path = self.cache[s3_path]
                return open(local_path, 'r' + 'b' if binary else '')
            except Exception as e:
                log.warn('S3 object {} not in cache: {}'.format(s3_path, e))

        local_path = key_download_decompress(self.bucket, self.prefix + s3_path)
        self.cache[s3_path] = local_path
        return open(local_path, 'r' + 'b' if binary else '')

    def put(self, source, name=None):
        fname = name if name else source.split('/')[-1]
        bucket = boto3.resource('s3').Bucket(self.bucket)
        if type(source) is str:
            bucket.upload_file(
                    self.prefix + fname, source)
            return True

        elif type(source) is io.BytesIO:
            bucket.upload_fileobj(
                    self.prefix + fname, source)
            return True

        else:
            raise Exception(
                "Can't put source data type {}. Pass BytesIO or string path to a local file.".format(type(source)))


class InMemoryConnector(base.Connector):
    def __init__(self, bucket, prefix='', binary=True, do_read=False,
                 credentials={}):
        super(InMemoryConnector, self).__init__(base_address=prefix, binary=binary, do_read=do_read)
        self.credentials = credentials
        self.bucket = bucket

    def get(self, name=None):
        """
        Get a bytes stream of an s3 key

        Parameters
        ----------
        name
            if None, prefix will be used as a key name, otherwise name will be added to prefix after '/' separator

        Returns
        -------
            bytes stream with key's data, or an empty stream if errors occurred
        """
        key = self.base_address if name is None else os.path.join(self.base_address, name)
        key = key.strip('/')

        try:
            res = boto3.resource('s3', **self.credentials).Bucket(self.bucket).Object(key).get()
            loglevel = log.warning if res['ContentLength'] == 0 else log.debug
            loglevel("Getting {} bytes from \tbucket={}\tkey={}".format(res['ContentLength'], self.bucket, key))
            body = res.pop('Body')
            self.metadata = res
        except Exception as e:
            log.error("{}\tbucket={}\tkey={}".format(e, self.bucket, key))
            if log.level == logging.DEBUG:
                raise e
            body = io.BytesIO()

        if self.do_read:
            body = body.read()

        if not self.binary:
            body = body.decode()

        return body

    def put(self, source, name=None):
        """

        Parameters
        ----------
        source
        name

        Returns
        -------

        """

        if isinstance(source, io.BytesIO):
            source = source.read()
        elif isinstance(source, (str, bytes)):
            pass
        else:
            log.error('Source should be either a string, or bytes or io.BytesIO, got {}'.format(type(source)))
            return False

        key = self.base_address if name is None else os.path.join(self.base_address, name)
        key = key.strip('/')
        try:
            res = boto3.resource('s3').Bucket(self.bucket).Object(key).put(Body=source)
            log.info("Uploaded {} bytes to \tbucket={}\tkey={}".format(len(source), self.bucket, key))
            return res['ResponseMetadata']['HTTPStatusCode'] == 200
        except Exception as e:
            log.error("{}\tbucket={}\tkey={}".format(e, self.bucket, key))
            return False



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['cranial', 'cranial.connectors', 'cranial.keyvalue']

package_data = \
{'': ['*'],
 'cranial': ['.mypy_cache/*',
             '.mypy_cache/3.5/*',
             '.mypy_cache/3.5/collections/*',
             '.mypy_cache/3.5/cranial/*',
             '.mypy_cache/3.5/cranial/__init__/*',
             '.mypy_cache/3.5/cranial/connectors/*',
             '.mypy_cache/3.5/importlib/*',
             '.mypy_cache/3.5/os/*',
             'datastore/*',
             'datastore/adapters/*'],
 'cranial.connectors': ['.mypy_cache/*',
                        '.mypy_cache/3.5/*',
                        '.mypy_cache/3.5/collections/*',
                        '.mypy_cache/3.5/cranial/*',
                        '.mypy_cache/3.5/cranial/connectors/*',
                        '.mypy_cache/3.5/cranial/connectors/s3/*',
                        '.mypy_cache/3.5/importlib/*',
                        '.mypy_cache/3.5/logging/*',
                        '.mypy_cache/3.5/os/*',
                        '.mypy_cache/3.5/unittest/*']}

install_requires = \
['boto3>=1.9,<2.0', 'cachetools>=2.1,<3.0', 'cranial-common>=0.2.0,<0.3.0']

setup_kwargs = {
    'name': 'cranial-datastore',
    'version': '0.3.3',
    'description': 'Data storage connectivity & utilities for Cranial.',
    'long_description': None,
    'author': 'Matt Chapman et al.',
    'author_email': 'Matt@NinjitsuWeb.com',
    'url': 'https://github.com/tribune/cranial-datastore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



import pickle
import json
from typing import Iterable
from cachetools import TTLCache

from cranial.common import logger
from cranial.datastore.dbapi import get_temp_db

log = logger.get(name='cranial.keyvalue')

# Default Databse Column type for values when blob is False
DEFAULT_TYPE = 'text'


class KeyValueDB(object):
    '''This object should mostly work like a dictionary, except it reads and
    writes from an external database. Subclasses can implement caching as
    appropriate.

    This class assumes that the database enforces uniqueness of the keys. I.e.,
    all SELECT queries will have LIMIT 1.

    >>> kv = get_mock()
    >>> kv['foo'] = 'a'
    >>> kv[1] = 'int'
    >>> kv.cache_clear()
    >>> kv['foo']
    'a'
    >>> kv[1]
    'int'
    >>> kv['bar'] = 'b'
    >>> kv.cache_clear()
    >>> kv['bar']
    'b'
    >>> kv['foo'] = 'c'
    >>> kv.cache_clear()
    >>> kv['foo'] = 'd'
    >>> kv.cache_clear()
    >>> kv['foo']
    'd'
    >>> kv['bar']
    'b'
    >>> kv.db.connection.close()
    '''

    def __init__(self, cursor,
                 table: str,
                 create_table: bool = False,
                 truncate_table: bool = False,
                 drop_table: bool = False,
                 keycol: str = 'key',
                 valcol: str = 'value',
                 blob=True,
                 keytype: str = 'text',
                 valtype: str = DEFAULT_TYPE,  # Default if `blob` is False.
                 placeholder: str = '?',
                 commit_fn: callable = None,
                 cachesize=2 ** 10,
                 cacheduration=3600):
        """Create a KeyValueDB object with  given connection and table.

        By default, the value type is pickled blob. Users can pass `blob=False`
        to serialize data as JSON in a text field, or `valtype='TYPE'` where
        TYPE is any type supported by the database.

        Parameters
        ----------
        cursor
        A DBAPI2 cursor, or at least an object with an .execute()
        method that accepts SQL query and a .fetchone() method that
        returns a row.

        table
        A table with 'key' and 'value' columns in the connected DB.

        create_table
        If True, execute 'create table if not exists ...'

        keycol
        Column name for key values, default 'key'

        valcol
        Column name for value values, default 'value'

        keytype
        If create_table, the datatype for the key. Default 'text'.

        valtype
        If create_table, the datatype for the value. Default 'blob' when
        `blob` parameter (below) is true, otherwise, 'text.'

        blob **DEPRECATED**
        If True, use pickle to (de)serialize, if False, use json. To support
        this, the cursor.execute() method must be able to accept a `bytes`
        parameter.

        placeholder
        The placeholder character required by the cursor's `paramstyle`.
        Defaults to '?'.

        commit_fn
        An optional function that will be called after writes with
        no arguments. If using a transactional DBAPI2 cursor without
        autocommit, this should probably be cursor.connection.commit
        unless you are handling transactions elsewhere.


        cachesize
        Max size for the LRU cache.

        cacheduration
        Seconds after which cache items will be expired.

        """
        self.db = cursor
        self.table = table
        self.keycol = keycol
        self.keytype = keytype
        self.valcol = valcol
        self.valtype = 'blob' if blob and valtype is DEFAULT_TYPE else valtype
        self.placeholder = placeholder
        self.commit = commit_fn
        self.cache = TTLCache(maxsize=cachesize, ttl=cacheduration)
        if create_table:
            self._create_table(drop_table)
        elif truncate_table:
            self.db.execute("TRUNCATE {};".format(table))

        self._get_table_columns()

    def _get_table_columns(self):
        """
        Not ideal way to look up what columns are there in the underlying table
        # @TODO: make this less dependent on database type
        """
        try:
            # Better for Cassandra/Scylla.
            self.db.execute("SELECT column_name FROM system_schema.columns "
                            "WHERE keyspace_name = 'default' and table_name = '{}'".format(self.table))
            self.table_columns = [r.column_name for r in self.db.fetchall()]
        except:
            log.info("Couldn't get columns through system_schema.")

            try:
                # For DBAPI2 databases.
                self.db.execute("SELECT * FROM {} LIMIT 1".format(self.table))
                self.table_columns = [d[0] for d in self.db.description]
            except:
                log.warning("Couldn't get columns through row.")

        if not self.table_columns:
            raise Exception("Could not initialize table_columns. Maybe table does not exist.")
        else:
            log.info("{} table has these columns: {}".format(self.table, self.table_columns))

        return self.table_columns

    def _create_table(self, drop_table):
        """
        helper method to run 'create table' query,
        optionally drop the table if it exists before creating it again
        Parameters
        ----------
        drop_table
            if True, try dropping table before creating
        """
        if drop_table:
            log.warning("Dropping table if exists: {}".format(self.table))
            self.db.execute(
                "DROP TABLE IF EXISTS {t} ".format(t=self.table))
            self.db.execute(
                """CREATE TABLE IF NOT EXISTS {t} (
                {k} {ktype} PRIMARY KEY,
                {v} {vtype});""".format(t=self.table,
                                        k=self.keycol,
                                        ktype=self.keytype,
                                        v=self.valcol,
                                        vtype=self.valtype))

    def _to_blob(self, v):
        return pickle.dumps(v, protocol=4) \
            if self.valtype is 'blob' else json.dumps(v)

    def __getitem__(self, key):
        """
        Basic get/set is tested throughout the class. Here we test the cache.

        >>> kv = get_mock()
        >>> kv['witch'] = 'duck'
        >>> 'witch' in kv.cache
        True
        """
        if not (key in self.cache):
            base_query = "SELECT {} from {} where {} = {} LIMIT 1".format(
                self.valcol, self.table, self.keycol, self.placeholder)
            self.db.execute(base_query, [key])
            result = self.db.fetchone()[0]
            unserialized = pickle.loads(result) \
                if self.valtype is 'blob' else json.loads(result)
            self.cache[key] = unserialized

        return self.cache[key]

    def get(self, key, default=None):
        """
        analog of dict.get

        Parameters
        ----------
        key
            key

        default
            if key is not found, return this value, default is None

        Returns
        -------
            value corresponding to the key, or default
        """
        try:
            result = self[key]
            if result is None:
                return default
        except KeyError as e:
            log.debug(str(e))
            return default
        except Exception as e:
            log.warning(str(e))
            return default

        return result

    def __setitem__(self, key, value):
        serialized_value = self._to_blob(value)
        # Subjectively, UPDATE||INSERT is faster than DEL+INSERT in Redshift.
        # This could possibly be optimized for Cassandra, it works fine.
        try:
            base_query = "UPDATE {} SET {} = {} WHERE {} = {}".format(
                self.table, self.valcol, self.placeholder, self.keycol,
                self.placeholder)
            self.db.execute(base_query, [serialized_value, key])
            if key in self.cache:
                del (self.cache[key])

            if self.db.rowcount < 1:
                base_query = "INSERT INTO {} ({}, {}) VALUES ({}, {})".format(
                    self.table, self.keycol, self.valcol,
                    self.placeholder, self.placeholder)
                self.db.execute(base_query, [key, serialized_value])
                self.cache[key] = value
            if self.commit:
                self.commit()
            return value
        except Exception as e:
            log.error('{}: {}'.format(type(e), e))
            log.error("Could not update {} => {}".format(key, value))

    def __delitem__(self, key):
        base_query = "DELETE from {} where {} = {}".format(
            self.table, self.keycol, self.placeholder)
        self.db.execute(base_query, [key])
        if key in self.cache:
            del (self.cache[key])
        if self.commit:
            self.commit()
        return key

    def __missing__(self, key):
        raise KeyError('No such key.')

    def cache_clear(self):
        """
        >>> kv = get_mock()
        >>> kv['witch'] = 'duck'
        >>> kv.cache_clear()
        >>> 'witch' in kv.cache
        False
        """
        self.cache = TTLCache(self.cache.maxsize, self.cache.ttl)

    def keys(self):
        self.db.execute("SELECT distinct {} from {}".format(
            self.keycol, self.table))
        for r in self.db.fetchall():
            yield r[0]

    def __len__(self):
        return self.db.execute("SELECT COUNT(1) from {}".format(
            self.table)).fetchone()[0]


class KeyValueReadOnly(KeyValueDB):
    """Access a Key-Value table with protection against writes.

    >>> kv = get_mock()
    >>> kv['foo'] = 'a'
    >>> kv[1] = 'int'
    >>> kv.cache_clear()
    >>> kv['foo']
    'a'
    >>> ro = KeyValueReadOnly(kv.db, kv.table, blob=False)
    >>> ro['foo']
    'a'
    >>> ro['foo'] = 'b'
    Traceback (most recent call last):
      File "keyvalue.py", line 000, in __setitem__
        raise Exception('Read-only data.')
    Exception: Read-only data.
    >>> del(ro['foo'])
    Traceback (most recent call last):
      File "keyvalue.py", line 000, in __delitem__
        raise Exception('Read-only data.')
    Exception: Read-only data.
    """

    def __init__(self, cursor, table: str,
                 keycol: str = 'key', valcol: str = 'value', blob=True,
                 keytype: str = 'text', valtype: str = DEFAULT_TYPE,
                 placeholder: str = '?'):
        super().__init__(cursor=cursor,
                         table=table,
                         keycol=keycol,
                         valcol=valcol,
                         blob=blob,
                         keytype=keytype,
                         valtype=valtype,
                         placeholder=placeholder)

    def __setitem__(self, key, value):
        raise Exception('Read-only data.')

    def __delitem__(self, key):
        raise Exception('Read-only data.')


""" Module Functions. """


class StringSerDe():
    def dumps(self, v):
        return str(v)

    def loads(self, v):
        # Return None as-is to indicate a Null column.
        return str(v) if v is not None else v


class BoolIntSerDe():
    def dumps(self, v):
        if v is True:
            return 1
        elif v is False:
            return 0
        else:
            return v

    def loads(self, v):
        if v is 1:
            return True
        elif v is 0:
            return False
        else:
            return v


class NoOpSerDe():
    def dumps(self, v):
        return v

    def loads(self, v):
        return v


class KeyRowDB(KeyValueDB):
    """ This should work well for Cassandra/Scylla, and OK on traditional
    Relational DBs with an index on the key. You Probably don't want to use this
    on Redshift, though performance should be acceptable if the key is your
    table's SORTKEY.

    Tests to demostrate interoperability with KeyValueDB class:

    >>> kv = get_mock()
    >>> kr = KeyRowDB(kv.db, kv.table)
    >>> kv['foo'] = 'a'
    >>> kr['foo']['value'] == 'a' and kr['foo']['other'] == None
    True
    >>> kr['foo'] = {'other': 'x', 'value': 'b'}
    >>> kv.cache_clear()
    >>> kv['foo']
    'b'
    >>> kr['foo'] = {'key': 'foo', 'other': 'x', 'value': 'b'}
    >>> kv['foo'] = 'c'
    >>> kr.cache_clear()
    >>> kr['foo']['value'] == 'c' and kr['foo']['other'] == 'x'
    True
    """

    def __init__(self, cursor,
                 table: str,
                 valcols: Iterable = None,
                 create_table: bool = False,
                 truncate_table: bool = False,
                 drop_table: bool = False,
                 keycol: str = 'key',
                 keytype: str = DEFAULT_TYPE,
                 valtype: str = DEFAULT_TYPE,
                 placeholder: str = '?',
                 commit_fn: callable = None,
                 blob=False,
                 serde: 'loads/dumps' = json,
                 cachesize=2 ** 10,
                 cacheduration=3600):
        """
        If table creation is desired, the user must provide valcols. By default,
        values will be stored in text columns as json. Alternatively, the user
        may set either `blob = True` to store pickled bytes, or set `serde` to
        an object that will convert values to/from strings.


        >>> cur = get_temp_db()
        >>> cur.execute("CREATE TABLE simple (key VARCHAR, knights VARCHAR)"
        ...   ).connection.commit()
        >>> simple = KeyRowDB(cur, 'simple')
        >>> simple['spam'] = {'knights': 'ni!'}
        >>> cur.execute("SELECT * FROM simple").fetchone()
        ('spam', '"ni!"')
        >>> dropped = KeyRowDB(cur, 'simple', valcols=['different'],
        ...  drop_table=True, create_table=True)
        >>> cur.execute("SELECT * FROM simple").fetchone() is None
        True
        >>> created = KeyRowDB(cur, 'created', valcols=['spam'],
        ...   create_table=True)
        >>> created['camelot'] = {'spam': 'alot'}
        >>> pickled = KeyRowDB(cur, 'pickled', valcols=['v'],
        ...  create_table=True, blob=True)
        >>> pickled['grenade'] = {'v': {'count': 5}}
        >>> pickled.cache_clear()
        >>> pickled['grenade']['v']['count'] is 5
        True
        >>> numeric = KeyRowDB(cur, 'numeric', keycol='num', keytype='int',
        ...   valcols= ['word'], create_table=True)
        >>> numeric[1] = {'word': 'foo'}
        >>> cur.execute("SELECT * FROM numeric").fetchone()
        (1, '"foo"')
        """
        assert callable(serde.loads)
        assert callable(serde.dumps)

        self.valcols = valcols
        super().__init__(cursor=cursor,
                         table=table,
                         keycol=keycol,
                         keytype=keytype,
                         valtype=valtype,
                         placeholder=placeholder,
                         blob=blob,
                         create_table=create_table,
                         truncate_table=truncate_table,
                         drop_table=drop_table,
                         commit_fn=commit_fn,
                         cachesize=cachesize,
                         cacheduration=cacheduration)

        del self.valcol
        if not create_table:
            existing_cols = set(self.table_columns)
            if self.valcols is not None:
                req_cols = set(self.valcols)
                assert all([c in existing_cols for c in req_cols]), \
                    "these columns do not exist in the already existing table: {}".format(req_cols - existing_cols)

        self.serde = pickle if self.valtype is 'blob' else serde

    def _check_coltypes(self, cols, coltype):
        """
        make sure that coltype is a dict where keys are column names and values are their data types

        Parameters
        ----------
        cols
            list of column names

        coltype
            - if dict, make sure all columns are in dict keys and fill with default data type if missing
            - if iterable, has to have the same number of items as column names,
                the two lists will be zipped together
            - if str, this type will be used for all columns

        Returns
        -------
            dict with {column:datatype}
        """
        if isinstance(coltype, dict):
            coltype = {c: coltype.get(c, DEFAULT_TYPE) for c in cols}
        elif isinstance(coltype, (list, tuple)):
            assert len(cols) == len(coltype), "when column types are given as list the " \
                                              "number should match the number of columns"
            coltype = dict(zip(cols, coltype))
        elif isinstance(coltype, str):
            coltype = {c: coltype for c in cols}
        else:
            raise Exception("column types should be either dict, list, tuple or string")
        return coltype

    def _create_table(self, drop_table):
        """
        >>> cur = get_temp_db()
        >>> serde = StringSerDe()
        >>> kr = KeyRowDB(cur, 'newtable', commit_fn=cur.connection.commit,
        ...               create_table = True,
        ...               valcols=['foo', 'bar' , 'baz'], serde=serde)
        >>> kr['spam'] = {'foo': 'a', 'bar': 'b'}
        >>> kr.cache_clear()
        >>> kr['spam'] == {'foo': 'a', 'bar': 'b', 'baz': None}
        True
        """
        if not self.valcols:
            raise Exception("Can't create table without defined columns.")
        if drop_table:
            log.warning("Dropping table if exists {}".format(self.table))
            self.db.execute(
                "DROP TABLE IF EXISTS {t} ".format(t=self.table))

        query = self._create_table_query()
        self.db.execute(query)
        log.info("Created table {} (if it didn't exist already)".format(
            self.table))

    def _create_table_query(self):
        """
        Factor out query definition and its execution from _create_table method
        """
        # make sure column types are set for every column
        self.valtype = self._check_coltypes(self.valcols, self.valtype)

        # Initialize columns with Key. Then append Additional columns.
        columns = ['{} {} PRIMARY KEY'.format(self.keycol, self.keytype)]
        columns += ['{} {}'.format(name, dtype)
                    for name, dtype in self.valtype.items()]
        query = "CREATE TABLE IF NOT EXISTS {table} ({columns});".format(
            table=self.table, columns=','.join(columns))
        return query

    def _execute_get_query(self, key):
        """
        Factor out the query definition and its execution from __getitem__.
        """
        # @TODO Update to support varying Column Types.
        base_query = "SELECT {} from {} where {} = {} LIMIT 1".format(
            '*' if not self.valcols else ','.join(self.valcols),
            self.table,
            self.keycol,
            self.placeholder)
        log.debug(base_query)
        self.db.execute(base_query, [key])

    def get(self, key, default=None):
        """

        Parameters
        ----------
        key
        default
            1. Can be a dict, with valcols as keys and values as default values
               for each column.
            2. Can be a list with the same number of values as the number of
               valcols, list items correspond
               to individual default values for each of the valcols.
            3. Can be anything else, in this case the same default value will
               be used for all columns.
            If there is a need to use default value of dict or list type for
            all columns - replicate the value and use option 1 or 2.

        Returns
        -------

        """
        if default is not None:
            for k in default:
                assert k in (self.valcols or self.table_columns)
        return super().get(key, default)

    def __getitem__(self, key):
        """

        Basic get/set is tested throughout the class. Here we test the cache.

        >>> cur = get_temp_db()
        >>> kr = KeyRowDB(cur, 'simple', valcols=['weight'], create_table=True)
        >>> kr['witch'] = {'weight': 'duck'}
        >>> 'witch' in kr.cache
        True
        """
        if not (key in self.cache):
            self._execute_get_query(key)
            columns = self.valcols or self.table_columns
            result = self.db.fetchone()
            if not result:
                raise KeyError("key {} is not found".format(key))
            # Pair values with column name.
            values_zip = zip(columns, result)

            # Deserialize.
            values = {}
            for k, v in values_zip:
                try:
                    v = self.serde.loads(v)
                except Exception as e:
                    # The Key column usually isn't serialized.
                    if v is not None and k != self.keycol:
                        log.warning('Exception for key {}: {}'.format(k, v))
                        log.warning(e)
                # If Deserializing failed, fall-through to using the column
                # data as-is.
                values[k] = v

            self.cache[key] = values

        return self.cache[key]

    def __setitem__(self, key, values: dict):
        assert type(values) is dict

        # make sure there is no keycol in values dict
        values.pop(self.keycol, None)
        data = [[f, self.serde.dumps(v)] for f, v in values.items()]

        try:
            values_str = ', '.join(
                ['{} = {}'.format(f, self.placeholder) for f, v in data])
            base_query = "UPDATE {} SET {}".format(self.table, values_str)
            base_query += " WHERE {} = {}".format(self.keycol, self.placeholder)
            log.debug(base_query)
            self.db.execute(base_query, [v for f, v in data] + [key])

            if self.db.rowcount < 1:
                data.append([self.keycol, key])
                base_query = "INSERT INTO {} ({}) VALUES ({})".format(
                    self.table,
                    ', '.join([f for f, v in data]),
                    ', '.join([self.placeholder for _ in data]))
                log.debug(base_query)
                self.db.execute(base_query, [v for f, v in data])

            self.cache[key] = values

            if self.commit:
                self.commit()
            return values
        except Exception as e:
            log.error('{}: {}'.format(type(e), e))
            log.error("Could not update {} => {}".format(key, values))


class CompoundKeyRowDB(KeyRowDB):
    """ This CompoundKeyRowDB object takes tuples as keys.
    @TODO It might be an improvement to eliminate this class and simply have
    the code here used in KeyRowDB when KeyRowDB.keycol is a tuple.
    """

    def __init__(self, cursor,
                 table: str,
                 keycols: Iterable = None,
                 valcols: Iterable = None,
                 create_table: bool = False,
                 truncate_table: bool = False,
                 drop_table: bool = False,
                 keytype = DEFAULT_TYPE,
                 valtype = DEFAULT_TYPE,
                 placeholder: str = '?',
                 commit_fn: callable = None,
                 blob=False,
                 # @TODO Define Serde Abstract Base Class.
                 serde: 'loads/dumps' = json,
                 cachesize=2 ** 10,
                 cacheduration=3600):
        """
        If table creation is desired, the user must provide valcols. By default,
        values will be stored in text columns as json. Alternatively, the user
        may set either `blob = True` to store pickled bytes, or set `serde` to
        an object that will convert values to/from strings.


        >>> cur = get_temp_db()
        >>> cur.execute("CREATE TABLE simple (a VARCHAR, b VARCHAR, c VARCHAR, d VARCHAR)"
        ...   ).connection.commit()

        >>> simple = CompoundKeyRowDB(cur, 'simple', keycols=['a', 'b'])
        >>> simple[('foo', 'bar')] = {'c': 'knights', 'd': 'ni!'}
        >>> cur.execute("SELECT * FROM simple").fetchone()
        ('foo', 'bar', '"knights"', '"ni!"')

        >>> dropped = CompoundKeyRowDB(cur, 'simple', keycols=['a', 'b'], valcols=['different'],
        ...  drop_table=True, create_table=True)
        >>> cur.execute("SELECT * FROM simple").fetchone() is None
        True

        >>> created = CompoundKeyRowDB(cur, 'created', keycols=['a', 'b'], valcols=['spam'],
        ...   create_table=True)
        >>> created[('cam', 'elot')] = {'spam': 'alot'}

        >>> pickled = CompoundKeyRowDB(cur, 'pickled', keycols=['a', 'b', 'c'], valcols=['v'],
        ...  create_table=True, blob=True)
        >>> pickled[('gre', 'n', 'ade')] = {'v': {'count': 5}}
        >>> pickled.cache_clear()
        >>> pickled[('gre', 'n', 'ade')]['v']['count'] is 5
        True

        >>> numeric = CompoundKeyRowDB(cur, 'numeric', keycols=['a', 'b'], keytype='int',
        ...   valcols=['word', 'num'], valtype=['text', 'decimal'], create_table=True)
        >>> numeric[(1, 2)] = {'word': 'foo', 'num': 3.14}
        >>> cur.execute("SELECT a, b, word, num FROM numeric").fetchone()
        (1, 2, '"foo"', 3.14)
        """
        assert callable(serde.loads)
        assert callable(serde.dumps)

        self.keycols = keycols
        super().__init__(cursor=cursor,
                         table=table,
                         valcols=valcols,
                         keytype=keytype,
                         valtype=valtype,
                         placeholder=placeholder,
                         blob=blob,
                         create_table=create_table,
                         truncate_table=truncate_table,
                         drop_table=drop_table,
                         commit_fn=commit_fn,
                         cachesize=cachesize,
                         cacheduration=cacheduration)
        del self.keycol

    def keys(self):
        raise NotImplementedError("This class does not have method 'keys'")

    def _create_table_query(self):
        """
        Different way to define columns and set compound primary key.
        This is used inside _create_table method

        >>> cur = get_temp_db()
        >>> serde = StringSerDe()
        >>> kr = CompoundKeyRowDB(cur, 'newtable', commit_fn=cur.connection.commit,
        ...               create_table = True, keycols=['a', 'b'],
        ...               valcols=['foo', 'bar' , 'baz'], serde=serde)
        >>> kr[('spam', 'alot')] = {'foo': 'a', 'bar': 'b'}
        >>> kr.cache_clear()
        >>> kr[('spam', 'alot')] == {'foo': 'a', 'bar': 'b', 'baz': None}
        True
        """
        # make sure column types are set for every column
        self.valtype = self._check_coltypes(self.valcols, self.valtype)
        self.keytype = self._check_coltypes(self.keycols, self.keytype)
        log.debug(self.valtype)
        log.debug(self.keytype)

        columns = []
        columns += ['{} {}'.format(name, dtype) for name, dtype in self.keytype.items()]
        columns += ['{} {}'.format(name, dtype) for name, dtype in self.valtype.items()]
        query = "CREATE TABLE IF NOT EXISTS {table} ({columns}, PRIMARY KEY ({key_cols}))".format(
            table=self.table, columns=','.join(columns), key_cols=','.join(self.keycols))
        return query

    def _execute_get_query(self, key):
        """
        Different select query to filter on multiple column.
        This is used inside __getitem__ method
        """
        # @TODO Update to support varying Column Types.
        base_query = "SELECT {} from {} where {} LIMIT 1".format(
            '*' if not self.valcols else ','.join(self.valcols),
            self.table,
            " AND ".join(["{} = {}".format(c, self.placeholder) for c in self.keycols])
        )
        log.debug(base_query)
        self.db.execute(base_query, key)

    def __setitem__(self, key, values: dict):
        """
        To adapt to the fact that key is a tuple:
        - different way to make sure there is no keycol in values dict
        - different UPDATE query and exec
        - different INSERT query and exec
        """
        assert type(values) is dict

        # make sure there is no keycol in values dict
        _ = [values.pop(c, None) for c in self.keycols]
        data = [[f, self.serde.dumps(v)] for f, v in values.items()]

        try:
            values_str = ', '.join(
                ['{} = {}'.format(f, self.placeholder) for f, v in data])
            base_query = "UPDATE {} SET {}".format(self.table, values_str)
            base_query += " WHERE " + " AND ".join(["{} = {}".format(c, self.placeholder) for c in self.keycols])
            log.debug(base_query)
            self.db.execute(base_query, [v for f, v in data] + list(key))

            if self.db.rowcount < 1:
                data.extend(zip(self.keycols, key))
                base_query = "INSERT INTO {} ({}) VALUES ({})".format(
                    self.table,
                    ', '.join([f for f, v in data]),
                    ', '.join([self.placeholder for _ in data]))
                log.debug(base_query)
                self.db.execute(base_query, [v for f, v in data])

            self.cache[key] = values

            if self.commit:
                self.commit()
            return values
        except Exception as e:
            log.error('{}: {}'.format(type(e), e))
            log.error("Could not update {} => {}".format(key, values))

    def __delitem__(self, key):
        """
        Override to adopt query for mylti-column key
        """
        base_query = "DELETE from {}".format(self.table)
        base_query += " WHERE " + " AND ".join(["{} = {}".format(c, self.placeholder) for c in self.keycols])
        log.debug(base_query)
        self.db.execute(base_query, key)
        if key in self.cache:
            del (self.cache[key])
        if self.commit:
            self.commit()
        return key


class StringKeyRowDB(KeyRowDB):
    """Forces keys to be strings. Mainly for SQLite & testing."""

    def __getitem__(self, key):
        return super().__getitem__(str(key))

    def __setitem__(self, key, values):
        return super().__setitem__(str(key), values)

    def __delitem__(self, key):
        return super().__delitem__(str(key))


class UncountableKeyRowDB(KeyRowDB):
    def __len__(self):
        return 1


def get_mock(name='test', filename=None):
    cur = get_temp_db(filename)
    cur.execute(
        "CREATE TABLE {} (key VARCHAR, value VARCHAR, other VARCHAR);".format(
            name)).connection.commit()
    return KeyValueDB(cur, name, blob=False, commit_fn=cur.connection.commit)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from cranial.keyvalue import keyvalue


class DeleteOp:
    """This class is passed as the 'value' to the intercept function for a
    __delitem__ operation. This allows the intercept function to differentiate
    a Delete from a Get."""
    pass

class GetOp:
    """ @see DeleteOp."""
    pass


class KeyValueInterceptor(keyvalue.KeyValueDB):
    ''' Adds to KeyValueDB the ability provide a "interceptor" function that can
    make modifications based on the key.

    >>> import pickle
    >>> kv = get_mock()
    >>> def fn(obj, key, value=None):
    ...     obj.valcol = 'alt' if key == 'bar' else 'value'
    ...     return key, value
    ...
    >>> kv.set_interceptor(fn)
    >>> kv['foo'] = 'a'
    >>> kv['foo']
    'a'
    >>> kv['bar'] = 'b'
    >>> kv['bar']
    'b'
    >>> list(map(lambda x: (x[0], pickle.loads(x[1]) if x[1] else x[1], pickle.loads(x[2]) if x[2] else x[2]), kv.db.execute("SELECT * FROM test").fetchall()))
    [('foo', 'a', None), ('bar', None, 'b')]
    >>> del(kv['foo'])
    >>> kv.db.connection.close()
    '''

    def set_interceptor(self, fn: callable):
        """ fn(self, key, value) -> key, value"""
        if hasattr(self, 'intercept'):
            raise Exception("Intercept function should only be set once.")
        else:
            self.intercept = fn

    def __getitem__(self, key):
        if hasattr(self, 'intercept'):
            key, _ = self.intercept(self, key, GetOp)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        if hasattr(self, 'intercept'):
            key, value = self.intercept(self, key, value)
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        if hasattr(self, 'intercept'):
            key, _ = self.intercept(self, key, DeleteOp)
        return super().__delitem__(key)


""" Module Functions. """


def get_mock():
    import sqlite3
    conn = sqlite3.connect(":memory:")
    cur = conn.cursor()
    cur.execute(
        "CREATE TABLE test (key VARCHAR, value BLOB, alt BLOB);"
    ).connection.commit()
    return KeyValueInterceptor(cur, 'test', commit_fn=conn.commit)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from datetime import datetime
import io
import os
from tempfile import mkstemp

from cranial.connectors import base
from cranial.common import logger

log = logger.get(name='local_fetchers')  # streaming log


def file_readlines(fp, delete_after=False):
    """
    memory efficient iterator to read lines from a file (readlines() method reads whole file)
    Parameters
    ----------
    fp
        path to a decompressed file downloaded from s3 key
    delete_after
        delete file after it was read
    Returns
    -------
        generator of lines
    """
    with open(fp) as f:
        while True:
            line = f.readline()
            if line:
                yield line
            else:
                break

    if delete_after:
        try:
            os.unlink(fp)
        except Exception as e:
            log.warning(e)


class Connector(base.Connector):
    def __init__(self, path: str = '', binary=True, do_read=False) -> None:
        super().__init__(base_address=path, binary=binary, do_read=do_read)
        self._open_files = []  # type: List[FileHandle]

    def get(self, name=None):
        if name.startswith('/'):
            name = name[1:]
        filepath = self.base_address if name is None else os.path.join(self.base_address, name)
        try:
            mode = 'rb' if self.binary else 'r'
            fh = open(filepath, mode)
            self._open_files.append(fh)
            res = fh
            self.metadata = {
                'LastModified': datetime.utcfromtimestamp(
                    os.path.getmtime(filepath))}
            log.info("Opened \t{}".format(filepath))

        except Exception as e:
            log.error("{}\tbase_address={}\tname={}\tfilepath={}".format(
                e, self.base_address, name, filepath))
            res = io.BytesIO() if self.binary else io.StringIO()

        if self.do_read:
            res = res.read()

        return res

    def put(self, source, name: str = None) -> bool:

        filepath = self.base_address if name is None else os.path.join(self.base_address, name)
        dir_path = os.path.split(filepath)[0]
        if len(dir_path) > 0:
            os.makedirs(dir_path, exist_ok=True)

        if isinstance(source, io.IOBase):
            source = source.read()
        elif isinstance(source, (str, bytes)):
            pass
        else:
            log.error('Source should be either a string, bytes or a readable buffer, got {}'.format(type(source)))
            return False

        try:
            mode = 'wb' if self.binary else 'w'

            # first write to a temp file
            _, local_path = self.get_tmp_file()
            with open(local_path, mode) as f:
                f.write(source)

            # then rename temp file to a proper name
            os.rename(local_path, filepath)
            log.info("wrote to \t{}".format(filepath))
            return True

        except Exception as e:
            log.error("{}\tbase_address={}\tname={}".format(e, self.base_address, name))
            return False

    def get_tmp_file(self):
        fd, local_path = mkstemp(dir=self.base_address)
        os.close(fd)
        return fd, local_path

    def __del__(self):
        [fh.close() for fh in self._open_files]


if __name__ == "__main__":
    import doctest
    doctest.testmod()
